class BaseError(Exception):
    def __init__(self, **kw):
        for k, v in kw.items():
            setattr(self, k, v)

    def to_dict(self):
        return {
            "msg": getattr(self, "msg", ""),
            "code": getattr(self, "code", None),
            "data": getattr(self, "data", {}),
        }


class NotFoundError(BaseError):
    code = 404


class SecurityViolation(BaseError):
    pass


class AccessDenied(BaseError):
    msg = "Access denied"


class ValidationError(BaseError):
    code = 400


class InvalidSessionError(BaseError):
    code = 401
    msg = "Invalid session"


class ConflictError(BaseError):
    code = 409
    msg = "Duplicate resource"



import json


class ReadOnlyCachedModel:
    """
    Read only cached model
    """

    connection = None
    ns = None
    key_fields = None

    @classmethod
    def prefix_key(cls, data):
        key = cls.ns
        for field in cls.key_fields:
            key += f":{data[field]}"
        return key

    @classmethod
    def get(cls, **data):
        key = cls.prefix_key(data)
        value = cls.connection.get(key)
        return json.loads(value) if value else None

    @classmethod
    def exists(cls, **data):
        key = cls.prefix_key(data)
        return cls.connection.exists(key)

    @classmethod
    def get_count(cls, **data):
        key = cls.prefix_key(data)
        return int(cls.connection.get(key) or 0)


class ReadWriteCachedModel(ReadOnlyCachedModel):
    """
    Read/Write cached model
    """

    timeout = None

    @classmethod
    def _get_matched_keys(cls, data):
        pattern = cls.ns
        for field in cls.key_fields:
            pattern += f':{data.get(field, "*")}'
        return cls.connection.keys(pattern)

    @classmethod
    def create(cls, **data):
        key = cls.prefix_key(data)
        cls.connection.set(key, json.dumps(data))
        if cls.timeout:
            cls.connection.expire(key, cls.timeout)

    @classmethod
    def create_lookup(cls, **data):
        key = cls.prefix_key(data)
        cls.connection.set(key, 1)
        if cls.timeout:
            cls.connection.expire(key, cls.timeout)

    @classmethod
    def create_counter(cls, starting=1, **data):
        key = cls.prefix_key(data)
        cls.connection.set(key, starting)
        if cls.timeout:
            cls.connection.expire(key, cls.timeout)

    @classmethod
    def update(cls, **data):
        # key = cls.prefix_key(data)
        # NOTE: keepttl works only with Redis 6.0+
        # cls.connection.set(key, json.dumps(data), keepttl=True)
        cls.create(**data)

    @classmethod
    def increment(cls, amount=1, **data):
        key = cls.prefix_key(data)
        cls.connection.incr(key, amount)

    @classmethod
    def delete(cls, **data):
        key = cls.prefix_key(data)
        cls.connection.delete(key)

    @classmethod
    def delete_all(cls, **data):
        keys = cls._get_matched_keys(data)
        if keys:
            cls.connection.delete(*keys)



from kombu import Queue
from celery import Celery as BaseCelery


class Celery(BaseCelery):
    def add_task_queues(self, *queues):
        self.conf.task_queues = tuple(Queue(queue) for queue in queues if queue)

    def set_default_queue(self, queue):
        self.conf.task_default_queue = queue

    def add_periodic_task(
        self,
        schedule,
        task,
        queue=None,
        autoretry_for=tuple(),
        max_retries=3,
        default_retry_delay=180,
    ):
        """
        queue:
            The celery queue to which the receiver should respond.
        task:
            The name of the task to execute.
        schedule:
            The frequency of execution.
            This can be the number of seconds as an integer, a timedelta, or a crontab.
            You can also define your own custom schedule types, by extending the
            interface of schedule.
        autoretry_for:
            A list/tuple of exception classes. By default, no exceptions will be
            autoretried.
        max_retries:
            Maximum number of retries before giving up. A value of None means task will
            retry forever. By default, this option is set to 3.
        default_retry_delay:
            Default time in seconds before a retry of the task should be executed.
            Can be either int or float. Default is a three minute delay.
        """
        super().add_periodic_task(
            schedule,
            self.task(
                task,
                autoretry_for=autoretry_for,
                max_retries=max_retries,
                default_retry_delay=default_retry_delay,
            )
            .s()
            .set(queue=queue),
        )

    def task_with_apply_async(self, countdown=0, retry=True, queue=None, **task_kwargs):
        """Executes function asynchronously.
        Not using apply_async means that the task will not be executed by a celery
        worker, but in the current process instead just like a normal synchronous
        function.
        countdown -- executes in specified no of seconds from now.
        retry -- configures retry behavior.
        """

        def wrap(f):
            task = self.task(f, **task_kwargs)

            def wrapped_f(*args, **kwargs):
                return task.apply_async(
                    args, kwargs, countdown=countdown, retry=retry, queue=queue
                )

            return wrapped_f

        return wrap

    def clear_all_tasks(self):
        """
        Removes all pending tasks
        """
        return self.control.purge()



import os
import requests
import time
import apphelpers.sessions as sessionslib

from .app.models import globalgroups, sitegroups

from converge import settings


class urls:
    base = "http://127.0.0.1:8000/"
    echo = base + "echo"
    echo_for_registered = base + "secure-echo"
    echo_for_groups = base + "echo-groups"
    echo_for_sitegroups = base + "sites/1/echo-groups"
    echo_for_custom_authorization = base + "custom-authorization-echo"


pid_path = "tests/run/app.pid"

sessiondb_conn = dict(
    host=settings.SESSIONSDB_HOST,
    port=settings.SESSIONSDB_PORT,
    password=settings.SESSIONSDB_PASSWD,
    db=settings.SESSIONSDB_NO,
)
sessionsdb = sessionslib.SessionDBHandler(sessiondb_conn)
sessionsdb.destroy_all()


def gunicorn_setup_module():  # not working
    if os.path.exists(pid_path):
        os.remove(pid_path)
    cmd = f"gunicorn tests.service:__hug_wsgi__  -p {pid_path} -D"
    os.system(cmd)
    for i in range(10):
        if os.path.exists(pid_path):
            time.sleep(2)
            break


def gunicorn_teardown_module():
    if os.path.exists(pid_path):
        cmd = f"kill -9 `cat {pid_path}`"
        os.system(cmd)


def test_get():
    word = "hello"
    url = urls.echo + "/" + word
    assert requests.get(url).json() == word


def test_get_params():
    word = "hello"
    url = urls.echo + "/" + word
    params = {"word": word}
    assert requests.get(url, params=params).json() == word


def test_get_multi_params():
    nums = [3, 5]
    url = urls.base + "add"
    params = {"nums": nums}
    assert requests.get(url, params=params).json() == sum(nums)


def test_post():
    word = "hello"
    url = urls.echo
    assert requests.post(url, json={"word": word}).json() == word


def test_echo_for_registered():
    word = "hello"
    headers = {"NoAuthorization": "Header"}
    url = urls.echo_for_registered + "/" + word
    resp = requests.get(url, headers=headers)
    assert resp.status_code == 401


def test_user_id():
    uid = 101
    d = dict(uid=uid, groups=[])
    sid = sessionsdb.create(**d)

    headers = {"Authorization": sid}

    word = "hello"
    url = urls.echo + "/" + word
    assert requests.get(url, headers=headers).json() == ("%s:%s" % (uid, word))

    url = urls.base + "me/uid"

    data = {"uid": None}
    resp = requests.post(url, json=data, headers=headers)
    assert resp.json() == uid

    data = {"uid": 1}  # invalid claim
    resp = requests.post(url, json=data, headers=headers)
    assert resp.json() == uid

    data = {"uid": uid}
    resp = requests.post(url, json=data, headers=headers)
    assert resp.json() == uid


def test_group_access():
    # 1. No group
    uid = 111
    groups = []
    d = dict(uid=uid, groups=groups)
    sid = sessionsdb.create(**d)
    url = urls.echo_for_groups

    headers = {"Authorization": sid}
    assert requests.get(url, headers=headers).status_code == 403

    # 2. Forbidden group
    uid = 112
    groups = [globalgroups.forbidden.value]
    d = dict(uid=uid, groups=groups)
    sid = sessionsdb.create(**d)
    url = urls.echo_for_groups

    headers = {"Authorization": sid}
    assert requests.get(url, headers=headers).status_code == 403

    # 3. Access group
    uid = 113
    groups = [globalgroups.privileged.value]
    d = dict(uid=uid, groups=groups)
    sid = sessionsdb.create(**d)

    headers = {"Authorization": sid}
    assert requests.get(url, headers=headers).status_code == 200

    # 4. Other groups
    uid = 112
    groups = [globalgroups.others.value]
    d = dict(uid=uid, groups=groups)
    sid = sessionsdb.create(**d)
    url = urls.echo_for_groups

    headers = {"Authorization": sid}
    assert requests.get(url, headers=headers).status_code == 403


def test_not_found():
    uid = 117
    d = dict(uid=uid, groups=[], site_groups={})
    sid = sessionsdb.create(**d)

    headers = {"Authorization": sid}

    url = urls.base + "snakes/viper"
    assert requests.get(url).status_code == 404

    url = urls.base + "sites/1/snakes/viper"
    assert requests.get(url, headers=headers).status_code == 404


def test_site_group_access():
    # 1. No group
    uid = 114
    groups = []
    site_groups = {}
    d = dict(uid=uid, groups=groups, site_groups=site_groups)
    sid = sessionsdb.create(**d)
    url = urls.echo_for_sitegroups

    headers = {"Authorization": sid}
    assert requests.get(url, headers=headers).status_code == 403

    # 2. Forbidden group
    uid = 115
    groups = [globalgroups.forbidden.value]
    site_groups = {1: [sitegroups.forbidden.value]}
    d = dict(uid=uid, groups=groups, site_groups=site_groups)
    sid = sessionsdb.create(**d)
    url = urls.echo_for_sitegroups

    headers = {"Authorization": sid}
    assert requests.get(url, headers=headers).status_code == 403

    # 3. Access group
    uid = 116
    groups = [globalgroups.privileged.value]
    site_groups = {1: [sitegroups.privileged.value]}
    d = dict(uid=uid, groups=groups, site_groups=site_groups)
    sid = sessionsdb.create(**d)

    headers = {"Authorization": sid}
    assert requests.get(url, headers=headers).status_code == 200


def test_request_access():
    url = urls.base + "request-and-body"
    req = requests.post(url, data={"z": 1}, headers={"testheader": "testheader-value"})
    resp = req.json()
    assert "testheader".upper() in resp["headers"]
    assert resp["body"] == {"z": "1"}


def test_raw_request():
    url = urls.base + "request-raw-body"
    req = requests.post(url, data={"z": 1}, headers={"testheader": "testheader-value"})
    resp = req.json()
    assert "testheader".upper() in resp["headers"]


def test_custom_authorization_access():
    uid = 111
    groups = []
    d = dict(uid=uid, groups=groups)
    sid = sessionsdb.create(**d)
    headers = {"Authorization": sid}

    url = urls.echo_for_custom_authorization + "/authorized"
    assert requests.get(url).status_code == 401

    url = urls.echo_for_custom_authorization + "/authorized"
    assert requests.get(url, headers=headers).status_code == 200

    url = urls.echo_for_custom_authorization + "/unauthorized"
    assert requests.get(url, headers=headers).status_code == 403



import html2text
import os
import smtplib
import ssl

from email.utils import formataddr
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from email.mime.image import MIMEImage
from email.mime.base import MIMEBase
from email import encoders

from converge import settings


def send_email(
    sender,
    recipients,
    subject,
    text=None,
    html=None,
    attachments=None,
    images=None,
    reply_to=None,
    bcc=None,
):
    """
    text: text message. If html is provided and not text, text will be auto generated
    html: html message
    images: list of cid and image paths.
        eg. [('logo', 'images/logo.png'), ('Bruce', 'images/bat.png')]

    sender: can be sender email string e.g. 'foo@example.com' or
    list/tuple sender name and email  ('Foo', 'foo@example.com')

    """
    assert any((text, html)), "please provide html or text"

    if html and not text:
        text = html2text.html2text(html)

    from_header = sender
    if isinstance(sender, (list, tuple)):
        from_header = formataddr(sender)

    msg = MIMEMultipart("alternative")
    msg["Subject"] = subject
    msg["From"] = from_header
    msg["To"] = ", ".join(recipients)

    if bcc:
        msg["bcc"] = ", ".join(bcc)
    if reply_to:
        msg.add_header("reply-to", reply_to)

    msg.attach(MIMEText(text, "plain", "utf-8"))
    if html:
        msg.attach(MIMEText(html, "html", "utf-8"))
    if images:
        for cid, img in images:
            img_part = MIMEImage(img)
            img_part.add_header("Content-ID", "<" + cid + ">")
            msg.attach(img_part)
    if attachments:
        for file_path in attachments:
            file_name = os.path.basename(file_path)
            file_part = MIMEBase("application", "octet-stream")
            attachment = open(file_path, "rb")
            file_part.set_payload((attachment).read())
            encoders.encode_base64(file_part)
            file_part.add_header(
                "Content-Disposition", f"attachment; filename= {file_name}"
            )
            msg.attach(file_part)

    context = ssl.create_default_context()

    with smtplib.SMTP(settings.SMTP_HOST, settings.SMTP_PORT, timeout=3) as s:
        if settings.SMTP_USERNAME:
            s.ehlo()
            s.starttls(context=context)
            s.login(settings.SMTP_USERNAME, settings.SMTP_KEY)

        s.send_message(msg=msg)



# -*- coding: utf-8 -*-
from collections import namedtuple

from nose.tools import raises

import apphelpers.sessions as sessionslib
import settings

from apphelpers.errors import InvalidSessionError


sessiondb_conn = dict(
    host=settings.SESSIONSDB_HOST,
    port=settings.SESSIONSDB_PORT,
    password=settings.SESSIONSDB_PASSWD,
    db=settings.SESSIONSDB_NO,
)
sessionsdb = sessionslib.SessionDBHandler(sessiondb_conn)


class data:
    pass


class state:
    pass


Session = namedtuple("Session", ["uid", "groups", "k", "v"])
data.session = Session(987651, ["admin", "member"], "foo", "bar")
data.session_email = "noone@example.com"

# uid, groups, k, v = 987651, ['admin', 'member'], 'foo', 'bar'


def test_create():
    d = dict(
        uid=data.session.uid,
        groups=data.session.groups,
        extras=dict(email=data.session_email),
    )
    sid = sessionsdb.create(**d)
    assert len(sid) == 43
    sid_new = sessionsdb.create(data.session.uid, data.session.groups)
    assert sid == sid_new
    state.sid = sid


def test_update():
    sid = state.sid
    k, v = data.session.k, data.session.v
    sessionsdb.update(sid, {k: v})
    d = sessionsdb.get(sid)
    assert d[k] == v
    assert d["email"] == data.session_email
    sessionsdb.remove_from_session(sid, [k])
    d = sessionsdb.get(sid)
    assert k not in d


def test_resync():
    sid = state.sid
    k, v = data.session.k, data.session.v
    sessionsdb.update(sid, {k: v})
    d = sessionsdb.get(sid)
    assert d[k] == v

    d = dict(
        uid=data.session.uid,
        groups=data.session.groups,
        extras=dict(email=data.session_email),
    )
    sessionsdb.resync(sid, d)
    d = sessionsdb.get(sid)
    assert k not in d


@raises(InvalidSessionError)
def test_delete():
    sessionsdb.destroy(state.sid)
    assert sessionsdb.get(state.sid)


def test_session_lookups():
    uids = range(10000, 10010)
    groups = ["grp1", "grp2"]
    for uid in uids:
        sid = sessionsdb.create(uid, groups)
        assert sessionsdb.sid2uidgroups(sid) == (uid, groups)
        sessionsdb.destroy(sid)
        raises(InvalidSessionError)(sessionsdb.get)(sid)



import os
from requests_oauthlib import OAuth2Session

try:
    from converge import settings
except:
    import settings

if not settings.G_REDIRECT_URI.startswith("https"):
    os.environ["OAUTHLIB_INSECURE_TRANSPORT"] = "True"


def fetch_info(access_token):
    session = OAuth2Session(token={"access_token": access_token})
    userinfo = session.get("https://www.googleapis.com/oauth2/v1/userinfo").json()
    return userinfo



import hug

from apphelpers.rest.hug import APIFactory


def echo(word):
    return {"word": word}


def setup_routes(factory):
    factory.get("/echo/{word}")(echo)
    factory.post("/echo")(echo)


def make_app():
    router = hug.route.API(__name__)

    api_factory = APIFactory(router)
    setup_routes(api_factory)


make_app()
# gunicorn tests.services.basic:__hug_wsgi__



import hug
import settings

from apphelpers.rest.hug import APIFactory


def echo(word):
    return {"word": word}


def protected_echo(word):
    return {"word": word}


protected_echo.login_required = True


def setup_routes(factory):
    factory.get("/echo/{word}")(echo)
    factory.post("/echo")(echo)
    factory.get("/p/echo/{word}")(echo)
    factory.post("/p/echo")(echo)


def make_app():
    router = hug.route.API(__name__)
    api_factory = APIFactory(router)
    sessiondb_conn = dict(
        host=settings.SESSIONSDB_HOST,
        port=settings.SESSIONSDB_PORT,
        password=settings.SESSIONSDB_PASSWD,
        db=settings.SESSIONSDB_NO,
    )
    api_factory.setup_session_db(sessiondb_conn)
    setup_routes(api_factory)


make_app()
# gunicorn tests.services.session:__hug_wsgi__



from enum import Enum
import hug
import hug.directives

from apphelpers.rest.hug import user_id

from tests.app.models import globalgroups, sitegroups


def echo(word, user: hug.directives.user = None):
    return "%s:%s" % (user.id, word) if user else word


def secure_echo(word, user: hug.directives.user = None):
    return "%s:%s" % (user.id, word) if user else word


secure_echo.login_required = True


def echo_groups(user: hug.directives.user = None):
    return user.groups


echo_groups.groups_required = [globalgroups.privileged.value]
echo_groups.groups_forbidden = [globalgroups.forbidden.value]


def add(nums: hug.types.multiple):
    return sum(int(x) for x in nums)


def get_my_uid(uid: user_id):
    return uid


get_my_uid.login_required = True


def get_snake(name):
    return None


get_snake.not_found_on_none = True


def get_secure_snake(site_id, name):
    return None


get_secure_snake.not_found_on_none = True
get_secure_snake.login_required = True


def secure_multisite_echo(word, user: hug.directives.user = None):
    return "%s:%s" % (user.id, word) if user else word


secure_echo.login_required = True


def echo_multisite_groups(site_id: int, user: hug.directives.user = None):
    return user.groups


echo_multisite_groups.groups_required = [sitegroups.privileged.value]
echo_multisite_groups.groups_forbidden = [sitegroups.forbidden.value]


def process_request(request, body):
    return {"body": body, "headers": request.headers}


def process_raw_request(request):
    return {"raw_body": request.stream.read().decode(), "headers": request.headers}


process_raw_request.not_found_on_none = True


def check_authorization(user, *args, **kw):
    return kw["word"] == "authorized"


def custom_authorization_echo(word):
    return word


custom_authorization_echo.authorizer = check_authorization


def setup_routes(factory):

    factory.get("/echo/{word}")(echo)
    factory.post("/echo")(echo)

    factory.get("/add")(add)

    factory.get("/secure-echo/{word}")(secure_echo)
    factory.get("/custom-authorization-echo/{word}")(custom_authorization_echo)
    factory.get("/echo-groups")(echo_groups)

    factory.post("/me/uid")(get_my_uid)

    factory.get("/snakes/{name}")(get_snake)

    factory.get("/sites/{site_id}/secure-echo/{word}")(secure_multisite_echo)
    factory.get("/sites/{site_id}/echo-groups")(echo_multisite_groups)
    factory.get("/sites/{site_id}/snakes/{name}")(get_secure_snake)

    factory.post("/request-and-body")(process_request)
    factory.post("/request-raw-body", parse_body=False)(process_raw_request)

    # ar_handlers = (None, arlib.create, None, arlib.get, arlib.update, None)
    # factory.map_resource('/resttest/', handlers=ar_handlers)



import secrets
import _pickle as pickle
import redis
import hug

from apphelpers.errors import InvalidSessionError


_SEP = ":"
session_key = ("session" + _SEP).__add__

rev_lookup_prefix = "uid" + _SEP
rev_lookup_key = lambda uid: rev_lookup_prefix + str(uid)


THIRTY_DAYS = 30 * 24 * 60 * 60


class SessionDBHandler:
    def __init__(self, rconn_params):
        """
        rconn_params: redis connection parameters
        """
        self.rconn = redis.Redis(**rconn_params)

    def create(
        self, uid="", groups=None, site_groups=None, extras=None, ttl=THIRTY_DAYS
    ):
        """
        groups: list
        site_groups: dict
        extras (dict): each key-value pair of extras get stored into hset
        """
        if uid:
            sid = self.uid2sid(uid)
            if sid:
                return sid

        sid = secrets.token_urlsafe()
        key = session_key(sid)

        if groups is None:
            groups = []
        session_dict = {"uid": uid, "groups": groups, "site_groups": site_groups}
        if extras:
            session_dict.update(extras)
        session = {k: pickle.dumps(v) for k, v in session_dict.items()}
        self.rconn.hmset(key, session)

        if uid:
            rev_key = rev_lookup_key(uid)
            self.rconn.setex(rev_key, value=sid, time=ttl)
        self.rconn.expire(key, ttl)
        return sid

    def exists(self, sid):
        return self.rconn.exists(session_key(sid))

    def get(self, sid, keys=[]):
        s_values = self.rconn.hgetall(session_key(sid))
        if not s_values:
            raise InvalidSessionError()
        session = {k.decode(): pickle.loads(v) for k, v in s_values.items()}
        if keys:
            session = {k: session.get(k, None) for k in keys}
        return session

    def get_attribute(self, sid, attribute):
        value = self.rconn.hget(session_key(sid), attribute)
        return pickle.loads(value) if value else None

    def uid2sid(self, uid):
        sid = self.rconn.get(rev_lookup_key(uid))
        return sid.decode() if sid else None

    def sid2uid(self, sid):
        session = self.get(sid, ["uid"])
        return session["uid"]

    def get_for(self, uid):
        sid = self.uid2sid(uid)
        return self.get(sid) if sid else None

    # Same default ttl as `create` function
    def extend_timeout(self, sid, ttl=THIRTY_DAYS):
        self.rconn.expire(session_key(sid), ttl)
        uid = self.sid2uid(sid)
        self.rconn.expire(rev_lookup_key(uid), ttl)

    def sid2uidgroups(self, sid):
        """
        => uid (int), groups (list)
        """
        session = self.get(sid, ["uid", "groups"])
        return session["uid"], session["groups"]

    def update(self, sid, keyvalues):
        sk = session_key(sid)
        keyvalues = {k: pickle.dumps(v) for k, v in list(keyvalues.items())}
        self.rconn.hmset(sk, keyvalues)

    def update_for(self, uid, keyvalues):
        sid = self.uid2sid(uid)
        return self.update(sid, keyvalues) if sid else None

    def update_attribute(self, sid, attribute, value):
        key = session_key(sid)
        self.rconn.hset(key, attribute, pickle.dumps(value))
        return True

    def resync(self, sid, keyvalues):
        removed_keys = list(self.get(sid).keys() - keyvalues.keys())
        self.remove_from_session(sid, removed_keys)
        self.update(sid, keyvalues)

    def resync_for(self, uid, keyvalues):
        keyvalues["uid"] = uid
        sid = self.uid2sid(uid)
        return self.resync(sid, keyvalues) if sid else None

    def remove_from_session(self, sid, keys):
        sk = session_key(sid)
        if keys:
            self.rconn.hdel(sk, *keys)
        return True

    def destroy(self, sid):
        uid = self.sid2uidgroups(sid)[0]
        sk = session_key(sid)
        self.rconn.delete(sk)
        self.rconn.delete(rev_lookup_key(uid))
        return True

    def destroy_for(self, uid):
        sid = self.uid2sid(uid)
        return self.destroy(sid) if sid else None

    def destroy_all(self):
        keys = self.rconn.keys(session_key("*"))
        if keys:
            self.rconn.delete(*keys)
        keys = self.rconn.keys(rev_lookup_prefix + "*")
        if keys:
            self.rconn.delete(*keys)


def whoami(user: hug.directives.user):
    return user.to_dict()


whoami.login_required = True



import os
import logging
from logging.handlers import RotatingFileHandler
from loguru import logger as loguru_logger
from converge import settings


def build_api_logger():
    """
    Builds multiprocess-safe (hence loguru over stdlib logger) API logger
    """
    level = settings.API_LOGGER.LEVEL
    handler = settings.API_LOGGER.FILEPATH
    if handler:  # Else log to sys.stderr by default
        rotation = settings.API_LOGGER.ROTATION
        retention = settings.API_LOGGER.RETENTION
        loguru_logger.add(
            handler,
            retention=retention,
            rotation=rotation,
            format="{time:YYYY-MM-DD HH:mm:ss} | {message}",
            enqueue=True,
            level=level,
        )
    return loguru_logger


def build_app_logger(name="app", logfile="app.log", debug=True):
    """
    General purpose application logger. Useful mainly for debugging
    """
    level = logging.DEBUG if settings.DEBUG else logging.INFO
    logdir = "logs"  # TODO: move this to settings
    if not os.path.exists(logdir):
        os.mkdir(logdir)
    logpath = os.path.join(logdir, logfile)
    maxBytes = 1024 * 1024 * 10
    handler = RotatingFileHandler(logpath, maxBytes=maxBytes, backupCount=100)
    handler.setLevel(level)
    formatter = logging.Formatter("[%(levelname)s] %(asctime)s: %(message)s")
    handler.setFormatter(formatter)
    logger = logging.getLogger(name)
    logger.addHandler(handler)
    logger.setLevel(level)
    return logger


api_logger = build_api_logger() if settings.API_LOGGER.ENABLED else None
app_logger = build_app_logger()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages

with open("README.rst") as readme_file:
    readme = readme_file.read()

with open("CHANGELOG.rst") as history_file:
    history = history_file.read()

requirements = ["loguru"]

setup_requirements = []

test_requirements = []

setup(
    author="Scroll Tech",
    author_email="tech@scrollstack.com",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ],
    description="Common helper libraries for Python Apps",
    install_requires=requirements,
    license="MIT license",
    long_description=readme + "\n\n" + history,
    include_package_data=True,
    keywords="apphelpers",
    name="apphelpers",
    packages=find_packages(exclude=["tests"]),
    setup_requires=setup_requirements,
    test_suite="tests",
    tests_require=test_requirements,
    url="https://github.com/scrolltech/apphelpers",
    version="0.31.1",
    zip_safe=False,
)



from dataclasses import dataclass, asdict
from falcon import HTTPUnauthorized, HTTPForbidden, HTTPNotFound
import hug
from hug.decorators import wraps

from apphelpers.db.peewee import dbtransaction
from apphelpers.errors import InvalidSessionError
from apphelpers.sessions import SessionDBHandler
from converge import settings

from apphelpers.loggers import api_logger

if settings.get("HONEYBADGER_API_KEY"):
    from honeybadger import Honeybadger
    from honeybadger.utils import filter_dict


def phony(f):
    return f


def raise_not_found_on_none(f):
    if getattr(f, "not_found_on_none", None) == True:

        @wraps(f)
        def wrapper(*ar, **kw):
            ret = f(*ar, **kw)
            if ret is None:
                raise HTTPNotFound("four o four")
            return ret

        return wrapper
    return f


def honeybadger_wrapper(hb):
    """
    wrapper that executes the function in a try/except
    If an exception occurs, it is first reported to Honeybadger
    """

    def wrapper(f):
        @wraps(f)
        def f_wrapped(*args, **kw):
            try:
                ret = f(*args, **kw)
            except Exception as e:
                try:
                    hb.notify(
                        e,
                        context={
                            "func": f.__name__,
                            "args": args,
                            "kwargs": filter_dict(kw, settings.HB_PARAM_FILTERS),
                        },
                    )
                finally:
                    raise e
            return ret

        return f_wrapped

    return wrapper


@hug.directive()
def user_id(default=None, request=None, **kwargs):
    return request.context["user"].id


@hug.directive()
def user_name(default=None, request=None, **kwargs):
    return request.context["user"].name


@hug.directive()
def user_email(default=None, request=None, **kwargs):
    return request.context["user"].email


@hug.directive()
def domain(default=None, request=None, **kwargs):
    return request.headers["HOST"]


@hug.directive()
def user_mobile(default=None, request=None, **kwargs):
    return request.context["user"].mobile


@dataclass
class User:
    sid: str = None
    id: int = None
    name: str = None
    groups: tuple = ()
    email: str = None
    mobile: str = None
    site_groups: dict = None

    def to_dict(self):
        return asdict(self)

    def __bool__(self):
        return bool(self.id)


def setup_strict_context_setter(sessions):
    def set_context(token):

        uid, groups, name, email, mobile, site_groups = None, [], "", None, None, {}

        if token:
            try:
                session = sessions.get(
                    token, ["uid", "name", "groups", "email", "mobile", "site_groups"]
                )
                uid, name, groups, email, mobile, site_groups = (
                    session["uid"],
                    session["name"],
                    session["groups"],
                    session["email"],
                    session["mobile"],
                    session["site_groups"],
                )
            except InvalidSessionError:
                raise HTTPUnauthorized("Invalid or expired session")

        return User(
            sid=token,
            id=uid,
            name=name,
            groups=groups,
            email=email,
            mobile=mobile,
            site_groups=site_groups,
        )

    return set_context


def setup_context_setter(sessions):
    def set_context(response, request, context, module):
        """
        Only sets context based on session.
        Does not raise any error
        """
        uid, groups, name, email, mobile, site_groups = None, [], "", None, None, {}
        token = request.get_header("Authorization")
        if token:
            try:
                session = sessions.get(
                    token, ["uid", "name", "groups", "email", "mobile", "site_groups"]
                )
                uid, name, groups, email, mobile, site_groups = (
                    session["uid"],
                    session["name"],
                    session["groups"],
                    session["email"],
                    session["mobile"],
                    session["site_groups"],
                )
                if api_logger:
                    api_logger.info("{} | {} | {}", uid, request.method, request.url)

            except InvalidSessionError:
                pass

        request.context["user"] = User(
            sid=token,
            id=uid,
            name=name,
            groups=groups,
            email=email,
            mobile=mobile,
            site_groups=site_groups,
        )

    return set_context


class APIFactory:
    def __init__(self, router, urls_prefix=""):
        self.router = router
        self.db_tr_wrapper = phony
        self.access_wrapper = phony
        self.secure_router = None
        self.multi_site_enabled = False
        self.site_identifier = None
        self.urls_prefix = urls_prefix
        self.honeybadger_wrapper = phony

    def enable_multi_site(self, site_identifier):
        self.multi_site_enabled = True
        self.site_identifier = site_identifier

    def setup_db_transaction(self, db):
        self.db_tr_wrapper = dbtransaction(db)

    def setup_honeybadger_monitoring(self):
        api_key = settings.HONEYBADGER_API_KEY
        if not api_key:
            print("Info: Honeybadger API KEY not found. Honeybadger not set")
            return

        print("Info: Setting up Honeybadger")
        hb = Honeybadger()
        hb.configure(api_key=api_key)
        self.honeybadger_wrapper = honeybadger_wrapper(hb)

    def setup_session_db(self, sessiondb_conn):
        """
        redis_conn_params: dict() with below keys
                           (host, port, password, db)
        """
        self.sessions = SessionDBHandler(sessiondb_conn)
        set_context = setup_context_setter(self.sessions)
        self.router = self.router.http(requires=set_context)
        set_context = setup_strict_context_setter(self.sessions)
        self.secure_router = self.router.http(
            requires=hug.authentication.token(set_context)
        )

        def access_wrapper(f):
            """
            This is the authentication + authorization part
            """
            login_required = getattr(f, "login_required", None)
            groups_required = getattr(f, "groups_required", None)
            groups_forbidden = getattr(f, "groups_forbidden", None)
            authorizer = getattr(f, "authorizer", None)

            if login_required or groups_required or groups_forbidden or authorizer:

                @wraps(f)
                def wrapper(request, *args, **kw):

                    user = request.context["user"]

                    # this is authentication part
                    if not user.id:
                        raise HTTPUnauthorized("Invalid or expired session")

                    # this is authorization part
                    groups = set(user.groups)

                    if groups_required and not groups.intersection(groups_required):
                        raise HTTPForbidden("Unauthorized access")

                    if groups_forbidden and groups.intersection(groups_forbidden):
                        raise HTTPForbidden("Unauthorized access")

                    if authorizer and not authorizer(user, *args, **kw):
                        raise HTTPForbidden("Unauthorized access")

                    return f(*args, **kw)

            else:
                wrapper = f

            return wrapper

        def multisite_access_wrapper(f):
            """
            This is the authentication + authorization part
            """
            login_required = getattr(f, "login_required", None)
            groups_required = getattr(f, "groups_required", None)
            groups_forbidden = getattr(f, "groups_forbidden", None)
            authorizer = getattr(f, "authorizer", None)

            if login_required or groups_required or groups_forbidden or authorizer:

                @wraps(f)
                def wrapper(request, *args, **kw):

                    user = request.context["user"]

                    # this is authentication part
                    if not user.id:
                        raise HTTPUnauthorized("Invalid or expired session")

                    # this is authorization part
                    groups = set(user.groups)
                    if self.site_identifier in kw:
                        site_id = int(kw[self.site_identifier])
                        groups = groups.union(user.site_groups.get(site_id, []))

                    if groups_required and not groups.intersection(groups_required):
                        raise HTTPForbidden("Unauthorized access")

                    if groups_forbidden and groups.intersection(groups_forbidden):
                        raise HTTPForbidden("Unauthorized access")

                    if authorizer and not authorizer(user, *args, **kw):
                        raise HTTPForbidden("Unauthorized access")

                    return f(*args, **kw)

            else:
                wrapper = f

            return wrapper

        self.access_wrapper = (
            multisite_access_wrapper if self.multi_site_enabled else access_wrapper
        )

    def choose_router(self, f):
        login_required = hasattr(f, "login_required") and f.login_required
        return self.secure_router if login_required else self.router

    def build(self, method, method_args, method_kw, f):
        print(
            f"{method_args[0]} [{method.__name__.upper()}] => {f.__module__}:{f.__name__}"
        )
        m = method(*method_args, **method_kw)
        f = self.access_wrapper(
            self.honeybadger_wrapper(self.db_tr_wrapper(raise_not_found_on_none(f)))
        )
        # NOTE: ^ wrapper ordering is important. access_wrapper needs request which
        # others don't. If access_wrapper comes late in the order it won't be passed
        # request parameter.
        return m(f)

    def get(self, path, *a, **k):
        def _wrapper(f):
            router = self.choose_router(f)
            args = (path if path.startswith("/") else (self.urls_prefix + path),) + a
            return self.build(router.get, args, k, f)

        return _wrapper

    def post(self, path, *a, **k):
        def _wrapper(f):
            router = self.choose_router(f)
            args = (path if path.startswith("/") else (self.urls_prefix + path),) + a
            return self.build(router.post, args, k, f)

        return _wrapper

    def put(self, path, *a, **k):
        def _wrapper(f):
            router = self.choose_router(f)
            args = (path if path.startswith("/") else (self.urls_prefix + path),) + a
            return self.build(router.put, args, k, f)

        return _wrapper

    def patch(self, path, *a, **k):
        def _wrapper(f):
            router = self.choose_router(f)
            args = (path if path.startswith("/") else (self.urls_prefix + path),) + a
            return self.build(router.patch, args, k, f)

        return _wrapper

    def delete(self, path, *a, **k):
        def _wrapper(f):
            router = self.choose_router(f)
            args = (path if path.startswith("/") else (self.urls_prefix + path),) + a
            return self.build(router.delete, args, k, f)

        return _wrapper

    def map_resource(self, collection_url, resource=None, handlers=None, id_field="id"):
        if resource:
            raise NotImplementedError("Resource not supported yet")

        resource_url = collection_url + "{" + id_field + "}"
        assert isinstance(handlers, (list, tuple)), "handlers should be list or tuple"
        (
            get_collection,
            add_resource,
            replace_resource,
            get_resource,
            update_resource,
            delete_resource,
        ) = handlers

        if get_collection:
            self.get(collection_url)(get_collection)
        if add_resource:
            self.post(collection_url)(add_resource)
        if replace_resource:
            self.put(resource_url)(replace_resource)
        if get_resource:
            # get_resource_wrapped = get_or_not_found(get_resource)
            self.get(resource_url)(get_resource)
        if update_resource:
            self.patch(resource_url)(update_resource)
        if delete_resource:
            self.delete(resource_url)(delete_resource)
